# -*- coding: UTF-8 -*-
#
#add_phys_props.py
# Richard Chopping and the Predictive Mineral Discovery Cooperative
#Research Centre (pmd*CRC), 2007.
#This software is freely available to pmd*CRC partners and sponsors.
#
#Correspondence should be made to:
#Richard Chopping
#Geoscience Australia
#GPO Box 378
#Canberra, ACT, 2601
#Phone +61 2 6249 9349
#Email richard.chopping@ga.gov.au
#
#Version: 2.0 (18/04/2007)
#
#Description:
#This python script takes an output timestep .vtk file from pmdPyRt
#and calculates physical properties for each node within the vtk
#
#
#Notes:
#* Assumes the chemical system file is called G10_01.csy
#* Assumes the naming convention of James Cleverley
#* Gocad TSurf conversion assumes that the X property in the vtk is the x
#  spatially, the y property is the z property spatially, and assigns a constant
#  y property for the output TSurf (assumes the section runs E-W)
#* Assumes that the mineralogy is specified per node (a requirement for the output
#  as a gocad TSurf, which specifies properties at nodes of triangles)
#
#Useage:
#python add_phys_props.py [pmdPyRt_timestep.vtk] [all|ascii|vtk|gocad]
#* pmdPyRt_timestep.vtk is the filename of the timestep to calculate properties
#  This is optional. If you don't supply it, then the program will ask you to
#  provide (via standard input) a filename for a vtk file.
#* all|ascii|vtk|gocad is an option to specify the filetypes that this script
#  should output. REQUIRED!
#  ascii - produces an ASCII csv file with the density, magsus and vp
#          calculated for each cell
#  vtk   - produces a new vtk file, with the contents of the original vtk file
#          as well as the new properties density, magsus and vp
#  gocad - produces a gocad TSurf containing the volume proportion of the minerals
#          from the input vtk file, and the density, magsus and vp
#  all   - produces all three of the above output options
#
#Example useage
#python add_phys_props.py original_model.vtk ascii,vtk
#  uses original_model.vtk as the input file, and produces both ascii csv and vtk
#  output
#python add_phys_props.py ascii,gocad
#  asks the user for the name of the input vtk file, and produces both ascii and
#  gocad representations of the results


import sys, csv, time
from csv import *
from property_calcs import *
#****
#Constants area
#****




#just in case we need a newline
#here we're assuming Windows.
newLine = "\r\n"

def main():
	lines = []

	if sys.argv[1].find(".") >= 0:
		lines,vtkFileName = open_vtk_as_lines(sys.argv[1])
		try:
			types = type_param_parse(sys.argv[2])
		except NameError:
			types = type_param_parse("all")

	else:
		try:
			types = type_param_parse(sys.argv[1])
		except NameError:
			types = type_param_parse("all")
		print "Model file name (including extension .vtk): ",
		line = sys.stdin.readline()
		vtkFileName = line.rstrip("\n")
		lines,vtkFileName = open_vtk_as_lines(vtkFileName)


	#turn the types into more meaningful variables
	output_ascii_grid, output_new_vtk, output_tsurf = types

	numPoints = num_points_in_vtk(lines)

	solid_minerals_starts = determine_solid_mineral_starts(lines, numPoints)

	model_suffix = find_model_suffix(vtkFileName)

	pretty_mineral_names = model_mineral_names(lines, solid_minerals_starts)

	model_time = determine_model_time(lines)

	#------ import chemistry file
	print "setting up chemistry lookup table...",
	chemistry_file_name = "G10_01.csy"
	molar_vols,molar_mass = csy_reader(chemistry_file_name)
	print "...chemistry LUT complete!"
	print ""

	#now we go through and we will get out all the data for each mineral
	print "adding mineral data...",

	#now we add to dictionaries
	point_mineralogy_mols = [{} for i in range(numPoints)]
	point_mineralogy_vol = [{} for i in range(numPoints)]
	i = 0
	for mineral_start in solid_minerals_starts:
		start = mineral_start + 2	#the first piece of data is on the 2nd line after the SCALARS line
		end = mineral_start + numPoints + 2	#and we count through numPoints
		mineral = pretty_mineral_names[i]	#the pretty_mineral_names list is in the order that the minerals are in
		for j in range(start, end):		#so we can use that to step through for each mineral, for each point
			mineral_mols = float(lines[j][0])	#get the number of moles for this point
			point_mineralogy_mols[j - start][mineral] = mineral_mols	#j - start is how far through we are
			temp_dict = {}
			temp_dict[mineral] = mineral_mols
			mineral_vol = calculate_volume_from_chemistry(temp_dict, molar_vols)	#use this to get us a mineralogy in volume proportion
			point_mineralogy_vol[j - start][mineral] = mineral_vol	#now we have a mineralogy for this point specified in volume proportion
		i = i + 1
	#done!

	for i in range(numPoints):
		point_mineralogy_vol[i] = scale_mineralogy(point_mineralogy_vol[i])	#for every cell in the model, specified by i
	print "... done adding all mineral data!"					#we use the scale_mineralogy function to make it so the total
	print ""									#proportion is 1

	#we also want to get the x, y, z info to use for check modelling and checking
	#points run from line #6 (index: 5) to numPoints + 6...	
	x,y,z = get_node_positions(lines, numPoints)
	
	#We now know the mineralogy for every cell, specified both as number of moles and volumetric proportion
	#Calculate the density of each cell, using the chemistry we know
	print "calculating the density of each cell...",
	density = [-1 for i in range(numPoints)]
	for i in range(numPoints):
		density[i] = calculate_density(mineralogy=point_mineralogy_mols[i], type=from_chemistry, molar_mass_db=molar_mass, molar_volume_db=molar_vols)
	print "...done!"
	print ""

	#Load in the magsus database & calculate magnetic susceptibility
	print "loading the magsus database...",
	magsus_db_file = "magsus.csv"
	magsus_db = import_single_property_db(magsus_db_file)
	print "...done!"
	print "calculating the magnetic susceptibility of each cell...",
	magsus = [calculate_magsus(mineralogy=point_mineralogy_vol[i], type=linear_average, magsus_db=magsus_db) for i in range(numPoints)]
	print "...done!"
	print ""

	#Load in the vp database & calculate vp
	print "loading the Vp database...",
	vp_db_file = "vp.csv"
	vp_db = import_single_property_db(vp_db_file)
	print "...done!"
	print "calculating the Vp of each cell...",
	vp = [calculate_seismic_velocity(point_mineralogy_vol[i], time_average, prop_vp, vp_db) for i in range(numPoints)]
	print "...done!"
	print ""

	if output_ascii_grid:
		#now we need to recombine data together!
		#output file name = phys_prop_output_time.csv
		outputFileName = "phys_prop_output_" + model_suffix + "_" + model_time + ".csv"
		outputPrettyString = "(" + outputFileName + ")"
		print "writing output CSV file", outputPrettyString, "..."
		#check that we're writing this out correctly
		writer = csv.writer(open(outputFileName, "wb"))
		
		#output the column names

		print "...outputting column names..."
		writer.writerow(["X", "Y", "Z", "Density", "Magsus", "Vp"])
		
		#go through each cell for each mineral and write (along with solid vol. info)
		print "... beginning to write out point and mineral data..."
		for i in range(numPoints):
			writer.writerow([x[i], y[i], z[i], density[i], magsus[i], vp[i]])
		print "... done!"
		print ""


	if output_new_vtk:
		#We want to write out a new vtk file, which is the old VTK file
		#and then with relevant physical property information appended
	
		#calculate an appropriate file name and open it
		outputVtkFileName = "phys_props_export_" + model_suffix + ".vtk"
		print "writing output VTK file (" + outputVtkFileName + ")..."

		out = open(outputVtkFileName, "wb")
	
		#for each line in the original vtk file...
		print "... writing original VTK file contents...",
		for line in lines:
			#for each 'item' in the line (space delim)
			for item in line:
				#write it out to the file
				out.write(item + " ")
			#now we write out a new line character
			out.write(newLine)
		#now we have the entire vtk file written out to the new file
		#start with a blank link
		out.write("newLine")
		print "... done!"

		print "... writing out the physical properties..."	
		#write out the density
		#start with the header
		out.write("SCALARS density float 1" + newLine)
		out.write("LOOKUP_TABLE default" + newLine)
		#for each density (in terms of cells), write it out as a new line
		for cell_density in density:
			out.write(str(cell_density) + newLine)
		#blank line for formatting
		out.write(newLine)

		#now write out magsus
		out.write("SCALARS magsus float 1" + newLine)
		out.write("LOOKUP_TABLE default" + newLine)
		#for each magsus (in terms of cells), write it out as a new line
		for magsus_val in magsus:
			out.write(str(magsus_val) + newLine)
		#blank line for formatting
		out.write(newLine)

		#now write out vp
		out.write("SCALARS vp float 1" + newLine)
		out.write("LOOKUP_TABLE default" + newLine)
		#for each vp (in terms of cells), write it out as a new line
		for vp_val in vp:
			out.write(str(vp_val) + newLine)
		#blank line for formatting
		out.write(newLine)
		out.close()
		print "... new vtk all done!"
		print ""

	if output_tsurf:
		outputTSurfName = "phys_prop_output_" + model_suffix + "_" + model_time
		outputTSurfFileName = outputTSurfName + ".ts"
		print "writing the gocad tsurf (" + outputTSurfFileName + "..."

		out = open(outputTSurfFileName, "wb")

		#finally, let's write out a tsurf.
		print "... writing gocad header...",
		out.write("GOCAD TSurf 1" + newLine)
		out.write("HEADER {" + newLine)
		out.write("name:" + outputTSurfName + newLine)
		out.write("}" + newLine)
		outString = "PROPERTIES "
		for mineral in pretty_mineral_names:
			outString = outString + mineral + " "
		outString = outString + "density magsus vp"
		out.write(outString + newLine)
		out.write("TFACE" + newLine)
		print "... done!"

		#headers written, now we put out the properties
		print "... writing all properties...",
		for i in range(numPoints):
			outString = "PVRTX "
			new_x,new_y,new_z = transform_points(x[i], y[i], z[i])
			outString = outString + str(i + 1) + " " + str(new_x) + " " + str(new_y) + " " + str(new_z) + " "
			for mineral in pretty_mineral_names:
				outString = outString + str(point_mineralogy_vol[i][mineral]) + " "
			outString = outString + str(density[i]) + " " + str(magsus[i]) + " " + str(vp[i]) + " "
			out.write(outString + newLine)
		print "... done!"

		#now we need to grab the triangulation stuff
		print "... writing triangulation information...",
		numTriangles = int(lines[numPoints + 7][1])
		start = numPoints + 8
		end = numPoints + numTriangles + 8
		for i in range(start, end):
			t1 = int(lines[i][1]) + 1	
			t2 = int(lines[i][2]) + 1
			t3 = int(lines[i][3]) + 1
			outString = "TRGL " + str(t1) + " " + str(t2) + " " + str(t3) + newLine
			out.write(outString)
		out.write("END" + newLine)
		print "... done!"
		print "... gocad tsurf done!"
		print ""

	print "add_phys_props.py all done!"

def open_vtk_as_lines(vtkFileName):
	lines = []
	try:
		f = file(vtkFileName, "rb")
	except:
		print "Model file name (including extension .vtk): ",
		line = sys.stdin.readline()
		vtkFileName = line.rstrip("\n")
		f = file(vtkFileName)
	print ""
	print "Reading in the data file...",

	reader = csv.reader(f, SpaceDelim)

	vtkLines = 0
	for row in reader:
		col = []
		for cols in row:
			col.append(cols)
		lines.append(col)
		vtkLines = vtkLines + 1
	print " ... done!\nRead", vtkLines, "lines from the VTK file"
	print ""
	return [lines, vtkFileName]


def type_param_parse(param_string):
	ascii_grid = False
	new_vtk = False
	gocad_tsurf = False
	param_string = param_string.lower()

	if param_string == "all":
		ascii_grid = True
		new_vtk = True
		gocad_tsurf = True
	elif param_string.find("ascii") >= 0:
		ascii_grid = True
	elif param_string.find("csv") >= 0:
		ascii_grid = True
	elif param_string.find("vtk") >= 0:
		new_vtk = True
	elif param_string.find("gocad") >= 0:
		gocad_tsurf = True
	elif param_string.find("tsurf") >= 0:
		gocad_tsurf = True

	return_value = [ascii_grid, new_vtk, gocad_tsurf]

	return return_value

def num_points_in_vtk(vtkLines):
	#5th line of the VTK file, 2nd column is the number of points
	numPointsLine = 4
	numPointsCol = 1

	print "determining the number of points in the VTK file...",
	numPoints = int(vtkLines[numPointsLine][numPointsCol])
	print "... done!"
	print "There are", numPoints, "points within the VTK file"

	return numPoints

def determine_solid_mineral_starts(vtkLines, numPoints):
	solid_minerals_starts = []

	for i in range(len(vtkLines)):
		if vtkLines[i] == []:
			pass
		elif vtkLines[i][0] == "SCALARS":
			if is_solid_mineral_line(vtkLines[i]):
				solid_minerals_starts.append(i)
			else:
				i = i + numPoints - 1
	return solid_minerals_starts

def determine_model_time(vtkLines):
	#find the solid volume information start line
	for i in range(len(vtkLines)):
		if vtkLines[i] == []:
			pass
		elif vtkLines[i][0] == "SCALARS":
			if is_solid_volume_line(vtkLines[i]):
				solid_vol_start = i

	#we want to know what time this model represents
	#we can find this as it's stored in the solid_volume line
	#trick is that the time starts the character after the =
	#and the final character occurs before the final _
	#in the string that also has the solid volume information
	print "calculating the time that this model represents...",
	yearsStart = vtkLines[solid_vol_start][1].find("=") + 1
	yearsEnd = vtkLines[solid_vol_start][1].rfind("_")
	modelTime = vtkLines[solid_vol_start][1][yearsStart:yearsEnd]
	print "...done! model is at time =", modelTime, "years"
	return modelTime
	
def find_model_suffix(vtkFileName):
	#another trick is that the models may be different
	#we can distinguish between the models by the TSx_y part of the
	#vtk file name
	#we expect the file name to be export_TSx_y.vtk
	#we want the TSx_y string
	print "finding model suffix...",
	suffix = vtkFileName[7:-4]
	print "...suffix found:", suffix, "... done!"	
	return suffix

def model_mineral_names(lines, solid_minerals_starts):
	#time to pretty up the column names
	print "producing pretty column names...",
	model_mineral_names = []
	for mineral_start in solid_minerals_starts:
		# pretty up the names
		model_mineral_names.append(pretty_mineral(lines[mineral_start][1]))
	print "... done!"
	print "\nThe minerals in this model are:"
	for mineral in model_mineral_names:
		print mineral
	print ""
	return model_mineral_names
	
def pretty_mineral(text):
	end = text.find("(s)")
	outString = ""
	for i in range(end - 1):	#we use -1 because we don't want to include an _
		outString = outString + text[i]
	return outString.lower()
	
def is_solid_mineral_line(vtkLine):
	if vtkLine[1].find("(s)") >= 0:
		return True
	else:
		return False
		
def is_solid_volume_line(vtkLine):
	if vtkLine[1].find("solid_volume") >= 0:
		return True
	else:
		return False

def transform_points(x, y, z):
	x = float(x)
	y = float(y)
	z = float(z)
	return [x, 0, y]

def get_node_positions(lines, numPoints):
	x = [lines[i][0] for i in range(5, numPoints + 5)]
	y = [lines[i][1] for i in range(5, numPoints + 5)]
	z = [lines[i][2] for i in range(5, numPoints + 5)]
	return [x,y,z]
		


if __name__ == "__main__":
	main()
