# -*- coding: UTF-8 -*-
#
#property_calcs.py
# Richard Chopping and the Predictive Mineral Discovery Cooperative
#Research Centre (pmd*CRC), 2007.
#This software is freely available to pmd*CRC partners and sponsors.
#
#Correspondence should be made to:
#Richard Chopping
#Geoscience Australia
#GPO Box 378
#Canberra, ACT, 2601
#Phone +61 2 6249 9349
#Email richard.chopping@ga.gov.au
#
#Version: 1.0 (18/04/2007)
#
#Description:
#This library provides some methods to estimate physical properties based
#  on mineralogy, and some supporting routines.
#In this version, there are no attempts to account for properties at
#  various temperatures and pressures, i.e. we are effectively producing
#  the properties of 'lab' samples, although the mineralogy may only be
#  stable at temperatures and pressures beyond that of 'lab' samples.
#The methods are mainly sourced either from fundamental chemistry, or 
#  from the following references:
#* Practical Handbook of Physical Properties of Rocks and Minerals,
#  (1989). Carmichael, R.S. ed, CRC Press, Boca Raton, Florida, 741p.
#  ISBN: 0-8493-3703-8
#* Ji, S., Wang, Q., Xia, B., (2002). Handbook of Seismic Properties of
#  Minerals, Rocks and Ores, Ecole Polytechnique de Montreal, Canada,
#  630p.
#* Connolly, J., (2007). Calculation of Seismic Velocities with
#  Perple_X: Data Preparation and Computational Algorithm.
#  http://www.perplex.ethz.ch/perplex_seismic_velocity_I.html.
#
#Notes:
#The Voigt-Reuss-Hill method of seismic velocity estimation has currently
#  not been implemented, as a compilation of suitable properties for
#  current purposes has not yet been derived. Expected in version 2.
#The use of linear averaging in log-log space for magnetic susceptibility
#  determination is unstable and should not be used.
#
#Testing:
#This library includes rudimentary unit testing methods. The tests
#  examine each calculation library, as well as database import and
#  export methods. The testing routines are be located in the testing
#  area of the program.
#To test, simply call the library as a python script:
#  python property_calcs.py
import sys, csv, math

#Classes for CSV stuff
class SpaceDelim(csv.excel):
	delimiter = " "
	skipinitialspace = True
class TabDelim(csv.excel):
	delimiter = "\t"
	skipinitialspace = True


#constants_area
#properties
prop_vp = 1
prop_vs = 2
prop_vpvs = 3
prop_all = 4

#type - shared
linear_average = 1

#type - density
from_chemistry = 2

#type - magnetic susceptibility
log_log_average = 2
werner_1945 = 3
mooney_bleifuss_1953 = 4
grant_west_1965 = 5
parasnis_1973 = 6

#type - seismic velocities
time_average = 2
connolly_vrh = 3

#keys for the csy import return value
csy_volume_key = 0
csy_mass_key = 1

###############################################################################
#calculate_density()
#
#Preconditions:
# * We have a series of constants - look in the constants_area
# * Mineralogy is a dictionary of mineral, proportion
#   If you're doing a linear average, proportion had better be a vol %
#   If you're doing it on chemistry, you MUST have proportion as MOLS!
# * type is an integer, set by the constants above
#   o linear_average = 1
#   o from_chemistry = 2
#   Default is linear_average
# * density_db (optional), is a dictionary of mineral, density pairs
#   This is needed if you want to calculate a linear averaged density!
# * molar_mass_db (optional), is a dictionary of mineral, molar mass pairs
#   This is needed if you want to calculate a density from chemistry!
# * molar_volume_db (optional), is a dictionary of mineral, molar volume pairs
#   This is needed if you want to calculate a density from chemistry!
#
#Postconditions:
#The method returns a float, representing the density of the sample represented by
#mineralogy, calculated by your choice of method.
#The method has failed if you get a return value of -1
def calculate_density(mineralogy={}, type=1, density_db={}, molar_mass_db={}, molar_volume_db={}):
	#just in case something goes wrong
	return_value = -1
	
	if type == linear_average:
		return_value = calculate_linear_average_density(mineralogy, density_db)

	if type == from_chemistry:
		mass = calculate_mass_from_chemistry(mineralogy, molar_mass_db)
		volume = calculate_volume_from_chemistry(mineralogy, molar_volume_db)
		#avoid a singularity which could cause your computer to quantum tunnel through the desk and break your foot
		if volume != 0:
			return_value = mass / volume
	return return_value

###############################################################################
#calculate_linear_average_density
#
#Calculates a density from a mineralogy and a database of densities for each mineral
#
#Preconditions
# * Mineralogy is a dictionary of mineral name, vol % pairs
# * density_db is a dictionary of mineral name, density pairs
#
#Postconditions
# * Returns -1 if everything falls apart
# * Won't return a density equal to zero.
# * Density will be returned otherwise
def calculate_linear_average_density(mineralogy={}, density_db={}):
	#just in case something goes wrong
	return_value = -1
	density = 0
	for mineral in mineralogy.keys():
		mineral_proportion = mineralogy[mineral]
		mineral = mineral.lower()
		density = mineral_proportion * density_db[mineral] + density
	if density != 0:
		return_value = density
	return return_value

###############################################################################
#calculate_mass_from_chemistry
#
#Calculates the mass of a given mineralogy.
#
#Preconditions
# * molar_mineralogy is a database of mineral, mols pairs
# * molar_mass_db is a database of mineral, molar mass pairs
#   can be derived from a chemical system file from UT2K
#
#Postconditions
# * The calculated mass is returned.
def calculate_mass_from_chemistry(molar_mineralogy, molar_mass_db):
	mass = 0
	for mineral in molar_mineralogy.keys():
		mineral_mols = molar_mineralogy[mineral]
		mineral = mineral.lower()
		mass = mineral_mols * molar_mass_db[mineral] + mass
	return mass

###############################################################################
#calculate_volume_from_chemistry
#
#Calculates the volume of a given mineralogy.
#
#Preconditions
# * molar_mineralogy is a database of mineral, mols pairs
# * molar_volume_db is a database of mineral, molar volume for that mineral pair
#   Can be derived from a CSY file derived from UT2K
def calculate_volume_from_chemistry(molar_mineralogy, molar_volume_db):
	volume = 0
	for mineral in molar_mineralogy.keys():
		mineral_mols = molar_mineralogy[mineral]
		mineral = mineral.lower()
		volume = mineral_mols * molar_volume_db[mineral] + volume
	return volume

###############################################################################		
#calculate_magsus
#
#Preconditions
# * We have a series of constants -- go check up in the constants_area
# * mineralogy is a dictionary of mineral, vol% pairs
# * magsus_db is a dictionary of mineral, magnetic susceptibility pairs
#Postconditions
# * Returns a magnetic susceptibility, calculated by the appropriate method
#   -1 if something has gone wrong in the calculation
def calculate_magsus(mineralogy={}, type=1, magsus_db={}):
	magsus = -1

	if type == linear_average:
		magsus = calculate_magsus_linear_linear(mineralogy, magsus_db)

	if type == log_log_average:
		magsus = calculate_magsus_log_log(mineralogy, magsus_db)

	if type == werner_1945:
		magsus = calculate_magsus_werner_1945(mineralogy, magsus_db)

	if type == mooney_bleifuss_1953:
		magsus = calculate_magsus_mooney_bleifuss_1953(mineralogy, magsus_db)

	if type == grant_west_1965:
		magsus = calculate_magsus_grant_west_1965(mineralogy, magsus_db)

	if type == parasnis_1973:
		magsus = calculate_magsus_parasnis_1973(mineralogy, magsus_db)

	return magsus

###############################################################################
#calculate_magsus_linear_linear
#
#Calculate magsus using linear averaging in linear (magsus) space - linear
#magnetic material space
#This is an acceptable assumption to calculate magsus from mineralogy,
#see Figure 27, Carmichael, R.S. (1989), Magnetic Properties of Minerals
#  and Rocks, p. 347. In Carmichael, R.S. (Ed), 1989, Practical Handbook of Physical Propertiers
#  of Rocks and Minerals, CRC Press, 741 p.
#
#Preconditions
# * mineralogy is a dictionary of mineral, volume proportion
# * magsus_db is a dictionary of mineral, magsus values
#Postconditions
# * The magnetic susceptibility will be returned, in whatever units are used in
#   magsus_db
#
def calculate_magsus_linear_linear(mineralogy={}, magsus_db={}):
	magsus = 0
	for mineral in mineralogy.keys():
		mineral_prop = mineralogy[mineral]
		mineral = mineral.lower()
		magsus = magsus_db[mineral] * mineral_prop + magsus
	return magsus

###############################################################################
#calculate_magsus_log_log
#
#Calculate magsus using linear averaging in log (magsus) - log (mag content) space
#Idea taken from Figure 23, Carmichael, R.S. (1989), Magnetic Properties of Minerals
#  and Rocks, p. 347. In Carmichael, R.S. (Ed), 1989, Practical Handbook of Physical Propertiers
#  of Rocks and Minerals, CRC Press, 741 p.
#
#Preconditions
# * mineralogy is a dictionary of mineral, volume proportion
# * magsus_db is a dictionary of mineral, magsus values
#Postconditions
# * The magnetic susceptibility will be returned, in whatever units are used in
#   magsus_db
#
#Notes
#This currently does not not work well when you have high concentrations material with a low
#magnetic susceptibility!
def calculate_magsus_log_log(mineralogy={}, magsus_db={}):
	magsus_log = 0
	for mineral in mineralogy.keys():
		mineral_prop = mineralogy[mineral]
		mineral = mineral.lower()
		if mineral_prop != 0 and magsus_db[mineral] != 0:
			magsus_log = math.log10(mineral_prop) * math.log10(magsus_db[mineral]) + magsus_log
	if magsus_log != 0:
		magsus = pow(10, magsus_log)	
	else:
		magsus = 0

	return magsus

###############################################################################
#calculate_magsus_werner_1945
#
#Calculate magnetic susceptibility using the formula of Werner (1945)
#  Werner, S. (1945). Determinations of the magnetic susceptibilities of
#  ores and rocks from Swedish iron ore deposits. Swedish Geological
#  Survey, 39, 1.
#Preconditions
# * mineralogy is a dictionary of mineral, volume proportion
# * magsus_db is a dictionary of mineral, magsus values
#Postconditions
# * The magnetic susceptibility will be returned, calculated in SI values
#
def calculate_magsus_werner_1945(mineralogy, magsus_db):
	N = 4 / 3 * math.pi
	magsus = 0
	for mineral in mineralogy.keys():
		mineral_prop = mineralogy[mineral]
		mineral = mineral.lower()
		magsus = magsus + (mineral_prop * magsus_db[mineral]) / (1 + N * (1 - math.pow(mineral_prop, 1/6)))
	return magsus

###############################################################################
#calculate_magsus_mooney_bleifuss_1953
#
#Calculate magnetic susceptibility using the formula of Mooney
#and Bleifuss (1953)
#  Mooney, H.M. and Bleifuss, R. (1953). Magnetic susceptibility
#  measurements in Minnesota. II. Analysis of field results, Geophysics
#  18, 383.
#Preconditions
# * mineralogy is a dictionary of mineral, volume proportion
# * magsus_db is a dictionary of mineral, magsus values
#Postconditions
# * The magnetic susceptibility will be returned, calculated in SI values
#
def calculate_magsus_mooney_bleifuss_1953(mineralogy, magsus_db):
	magsus = 0
	for mineral in mineralogy.keys():
		mineral_prop = mineralogy[mineral]
		magsus = magsus + 0.289 * math.pow(mineral_prop, 1.01)
	return magsus

###############################################################################
#calculate_magsus_grant_west_1965
#
#Calculate magnetic susceptibility using the formula of Grant
#and West (1965)
#  Grant, F.S. and West, G.F. (1965). Interpretation theory in applied
#  geophysics, McGraw-Hill, New York.
#Preconditions
# * mineralogy is a dictionary of mineral, volume proportion
# * magsus_db is a dictionary of mineral, magsus values
#Postconditions
# * The magnetic susceptibility will be returned, calculated in SI values
#
def calculate_magsus_grant_west_1965(mineralogy, magsus_db):
	magsus = 0
	for mineral in mineralogy.keys():
		mineral_prop = mineralogy[mineral]
		magsus = magsus + 0.26 * math.pow(mineral_prop, 1.11)		
	return magsus

###############################################################################
#calculate_magsus_parasnis_1973
#
#Calculate magnetic susceptibility using the formula of Parasnis (1973)
#  Parasnis, D.S. (1973). Mining geophysics, revised ed., Elsevier,
#  Amsterdam.
#Preconditions
# * mineralogy is a dictionary of mineral, volume proportion
# * magsus_db is a dictionary of mineral, magsus values
#Postconditions
# * The magnetic susceptibility will be returned, in the units as used in magsus_db
#
def calculate_magsus_parasnis_1973(mineralogy, magsus_db):
	magsus = 0
	for mineral in mineralogy.keys():
		mineral_prop = mineralogy[mineral]
		magsus = magsus + 0.116 * math.pow(mineral_prop, 1.39)
	return magsus
		
###############################################################################
#calculate_seismic_velocity()
#
#Preconditions:
# * We have a series of constants -- go check up in the constants_area
# * mineralogy is a dictionary of mineral, vol% pairs
# * type is an integer, set by the constants up above
# * properties is an integer, set by the constants up above
# * vp_database (optional), is a dictionary of mineralogy, vp pairs
#   This is needed for calculating vp, vpvs, or all,
#   unless you are calculating by VRH!
# * vs_database (optional), is a dictionary of mineralogy, vp pairs
#   This is needed for calculating vp, vpvs, or all,
#   unless you are calculating by VRH!
# * vrh_database (optional), is a dictionary of mineralogy, [mineralogy_parameters list] pairs
#   This is needed if you wish to calculate by VRH!
# * Method also requires access to the following other methods:
#   o seismo_linear_average() if you wish to calculate via linear average
#   o seismo_time_average() if you wish to calculate by time (Backus) average
#   o seismo_connolly_vrh() if you wish to calculate using Connolly's VRH implementation
#
#Postconditions
#The method either returns a float, representing either vp, vs or vp/vs ratio
#(depending on desired calculation type), or it returns a tuple of vp, vs, vp/vs
#if you require all properties
#
#Exceptions
#chucks a wobbly if you feed it an unsupported calculation type
def calculate_seismic_velocity(mineralogy={}, type=1, properties=1, vp_database={}, vs_database={}, vrh_database={}):
	#set the return value just in case something goes wrong:
	return_value = -1

	#calculate vp (if needed)
	if properties == prop_vp or properties == prop_vpvs or properties == prop_all:
		#use the appropriate method
		if type == linear_average:
			vp = seismo_linear_average(mineralogy, vp_database)
		elif type == time_average:
			vp = seismo_time_average(mineralogy, vp_database)
		elif type == connolly_vrh:
			#do nothing
			pass
		else:
			#throw some kind of exception here
			pass

		#if we're done with it, we can set the return value here
		if properties == prop_vp:
			return_val = vp

	#calculate vs (if needed)
	if properties == prop_vs or properties == prop_vpvs or properties == prop_all:
		#use the appropriate method
		if type == linear_average:
			vs = seismo_linear_average(mineralogy, vs_database)
		elif type == time_average:
			vs = seismo_time_average(mineralogy, vs_database)
		elif type == connolly_vrh:
			#do nothing
			pass
		else:
			#throw some kind of exception here
			pass

		#if we're done with it, we can set the return value here
		if properties == prop_vs:
			return_val = vs

	#calculate vp/vs (if needed)
	if properties == prop_vpvs or properties == prop_all:
		vpvs = vp/vs
		#if we're done with it, we can set the return value here
		if properties == prop_vpvs:
			return_val = vpvs

	#create the tuple to return, if we need to
	if properties == prop_all:
		return_val = []
		return_val.append(vp)
		return_val.append(vs)
		return_val.append(vpvs)
	
	#the return area
	return return_val

###############################################################################
#seismo_linear_average
#
#Preconditions:
# * We have a series of constants -- go check up in the constants_area
# * mineralogy is a dictionary of mineral, vol% pairs
# * seismo_database is a dictionary of mineral, property pairs. This is whatever
# property you wish to calculate by linear average
#
#Postconditions
#The method either returns a float, representing the linear average of the velocities
#
#Exceptions
#chucks a wobbly if you don't have the required minerals in your seismo_database
def seismo_linear_average(mineralogy, seismo_database):
	property = 0
	for mineral in mineralogy.keys():
		vol_perc = mineralogy[mineral]
		#lower case to compare to the seismic velocity database!
		mineral = mineral.lower()
		mineral_property = seismo_database[mineral]
		property = property + (vol_perc * mineral_property)
	return property

###############################################################################
#seismo_time_average
#
#Uses Backus' averaging method
#Effectively this is:
#(velocity)^1 = volprop1 / mineral_velocity_1 + volprop2 / mineral_velocity_2 + ...
#etc
#
#Preconditions:
# * We have a series of constants -- go check up in the constants_area
# * mineralogy is a dictionary of mineral, vol% pairs
# * seismo_database is a dictionary of mineral, property pairs. This is whatever
# property you wish to calculate by linear average
#
#Postconditions
#The method either returns a float, representing the time average of the velocities
#
#Exceptions
#chucks a wobbly if you don't have the required minerals in your seismo_database
def seismo_time_average(mineralogy, seismo_database):
	property = 0
	for mineral in mineralogy.keys():
		vol_perc = mineralogy[mineral]
		#convert to lower case to make database-lithology comparisons simpler
		mineral = mineral.lower()
		mineral_property = seismo_database[mineral]
		property = vol_perc / mineral_property + property
	property = 1/property
	return property

def seismo_connolly_vrh(mineralogy, properties, vrh_database):
	#do connolly's method
	pass

###############################################################################
#import_single_property_db
#
#This is the method to import a database of a single property.
#Pretty simple, given a filename, with or without path, it opens a file
#and spits back a dictionary of mineral, property pairs.
#
#Preconditions
#* Requires the csv module to have been imported
#* Expects a dialect called TabDelim to have been created
#* Database format is "mineral nameTABsome_property"
#
#Postconditions
#* Database file is closed
#* A dictionary of mineral, velocity pairs has been returned.
def import_single_property_db(db_name, DelimType = csv.excel):
	prop_db = {}
	db = open(db_name, "rb")
	reader = csv.reader(db, DelimType)
	for mineral, property in reader:
		prop_db[mineral.lower()] = float(property)
	return prop_db

###############################################################################
#csy_reader
#
#This is a way of reading in the CSY file
#and is to be used until we handle the proper pmdPyRT
#method of storing the abstracted chemical system
#Note we store the mineral names internally in lower case
#makes comparisons easier.
#
#Preconditions
#* Requires a filename to be fed to it (name will do, with a path is even better)
#* Requires CSV to have been imported, and a SpaceDelim dialect created
#* Needs some constants (see top of file)
#
#Postconditions
#* Spits out a list of dictionaries
#  o First dictionary is the molar volume dictionary
#  o Second dictionary is the molar mass dictionary
#  o Each output dictionary is a list of mineral, property pairs
def csy_reader(filename):
	chem_file = open(filename, "rb")
	csy_as_csv = csv.reader(chem_file, SpaceDelim)
	chem_lines = []
	for line in csy_as_csv:
		chem_lines.append(line)
		
	#now find the first line that contains mineral
	mineral_line = 0
	for i in range(len(chem_lines)):
		if chem_lines[i][1] == "Minerals":
			mineral_line = i			

	molar_vols = {}
	molar_mass = {}
	
	for i in range(mineral_line + 1, len(chem_lines) - 1):
		#two ways this can go: trailing space or not
		#rather than munt the chem file more, we'll handle both cases:
		if chem_lines[i][-1] == "":
			key = chem_lines[i][-2].lower()
			mass = float(chem_lines[i][-3])
			vol = float(chem_lines[i][-4])
		else:
			key = chem_lines[i][-1].lower()
			mass = float(chem_lines[i][-2])
			vol = float(chem_lines[i][-3])
		molar_vols[key] = vol
		molar_mass[key] = mass

	result = []
	result.append(molar_vols)
	result.append(molar_mass)
	return result

###############################################################################
#scale_mineralogy
#
#Scales the mineralogy to total 1
#i.e. represent mineralogy as proportion
#
#Preconditions
# * 
#Postconditions
# *
#
def scale_mineralogy(mineralogy={}):
	new_mineralogy = {}
	total_prop = 0
	for mineral in mineralogy.keys():
		total_prop = mineralogy[mineral] + total_prop
	for mineral in mineralogy.keys():
		new_mineralogy[mineral] = mineralogy[mineral] / total_prop
	return new_mineralogy







####################### testing area ##########################################
# EVERYTHING BELOW HERE IS FOR THE PURPOSES OF SELF TESTING!!!
# YES, THAT'S RIGHT, NOTHING BELOW HERE IS IMPORTANT FOR THE ACTUAL PROPERTIES

#Testing thing
def result_string(type, val1, val2):
	result = "FAIL!"
	tolerance = 0.0000001	#we want errors no more than 1ppm!
	diff = abs(val2 - val1)
	if diff <= tolerance:
		result = "PASS!"
	return_string = type + result + " (" + str(val1) + " vs. " + str(val2) + ")"
	return return_string

#more testing thing
def result_string_for_lists(type, l1, l2):
	result = "PASS!"
	tolerance = 0.0000001	#we want errors no more than 1ppm!
	if len(l1) == len(l2):
		diff = []
		for index in range(len(l1)):
			diff.append(abs(l1[index] - l2[index]))
		for difference in diff:
			if difference > tolerance:
				result = "FAIL!"
				break
	else:
		result = "fail!"
	return_string = type + result + "\n(" + str(l1) + " vs. " + str(l2) + ")"
	return return_string

#another testing thing
#write out a dummy CSY file
def write_out_test_csy(filename):
	csy_out = open(filename, "wb")
	lines = []
	lines.append("# Minerals  ")
	lines.append("  6.    0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   2.00   0.00   0.00   0.00   1.00   0.00    -    22.688      60.084   Quartz          ")
	lines.append(" 10.    0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   3.00   0.00   0.00   0.00   0.00   0.00   4.00   0.00   0.00   0.00   0.00   0.00    -    44.524     231.539   Magnetite       ")
	lines.append(" 16.    0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.88   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   1.00   0.00   0.00    -    18.200      81.044   \"Pyrrhotite (Magnetic)\"    ")
	lines.append(" 17.    0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   1.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   2.00   0.00   0.00    -    23.940     119.979   Pyrite        ")
	lines.append(" 64.    0.00   3.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   2.00   1.00   0.00   0.00   0.00  12.00   0.00   0.00   0.00   3.00   0.00    -   140.870     398.308   \"Muscovite (Parallel)\"         ")
	lines.append(" 76.    0.00   1.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   0.00   1.00   0.00   0.00   0.00   8.00   0.00   0.00   0.00   3.00   0.00    -   108.690     278.332   K-Feldspar     ")
	lines.append("# End of file  ")
	for line in lines:
		csy_out.write(line + "\r\n")
	csy_out.close()

#YATT
#write out a database for a single property dictionary
#produces a tab-delimited file
def write_out_single_property_db(properties, filename):
	db = open(filename, "wb")
	writer = csv.writer(db, TabDelim)
	for mineral in properties.keys():
		outRow = []
		outRow.append(mineral.lower())
		outRow.append(properties[mineral])
		writer.writerow(outRow)
	db.close()
	#end it all!


#Self test area
#First we create a mineralogy for a sample
#We then write out some database files, to make life easier
#We can read these database files using the appropriate readers
#We calculate properties from these database files,
#both within this method and using the 'real' callable methods
#We compare the results
def main():
	import os

	print "Self test of:", sys.argv[0]

	#SET UP AREA
	
	#set up appropriate proportions
	qtz_proportion = 0.45
	py_proportion = 0.15
	pomag_proportion = 0.10
	kspar_proportion = 0.15
	mgt_proportion = 0.05
	msvtpar_proportion = 0.10

	#add minerals to the bulk mineralogy of the sample
	mineralogy = {}
	mineralogy["Quartz"] = qtz_proportion
	mineralogy["Pyrite"] = py_proportion
	mineralogy["Pyrrhotite (Magnetic)"] = pomag_proportion
	mineralogy["K-Feldspar"] = kspar_proportion
	mineralogy["Magnetite"] = mgt_proportion
	mineralogy["Muscovite (Parallel)"] = msvtpar_proportion
	
	#Create a density database for linear averaging of density
	qtz_den = 2.650
	py_den = 5.040
	pomag_den = 4.620
	kspar_den = 2.540
	mgt_den = 5.125
	msvtpar_den = 2.810
	den_dict = {}
	den_dict["Quartz"] = qtz_den
	den_dict["Pyrite"] = py_den
	den_dict["Pyrrhotite (Magnetic)"] = pomag_den
	den_dict["K-Feldspar"] = kspar_den
	den_dict["Muscovite (Parallel)"] = msvtpar_den
	den_dict["Magnetite"] = mgt_den

	#Create a magsus database for testing the magnetic susceptibility work
	#for convenience, we're doing this is x 10^-3 units (SI)
	qtz_magsus = 0
	py_magsus = 0.15
	pomag_magsus = 150
	kspar_magsus = 0.00001
	mgt_magsus = 1500
	msvtpar_magsus = 0.0001
	magsus_dict = {}
	magsus_dict["Quartz"] = qtz_magsus
	magsus_dict["Pyrite"] = py_magsus
	magsus_dict["Pyrrhotite (Magnetic)"] = pomag_magsus
	magsus_dict["K-Feldspar"] = kspar_magsus
	magsus_dict["Muscovite (Parallel)"] = msvtpar_magsus
	magsus_dict["Magnetite"] = mgt_magsus

	#Create a Vp database
	qtz_vp = 6150
	py_vp = 8000
	pomag_vp = 4645
	kspar_vp = 6600
	mgt_vp = 7350
	msvtpar_vp = 4500
	vp_dict = {}
	vp_dict["Pyrrhotite (Magnetic)"] = pomag_vp
	vp_dict["Quartz"] = qtz_vp
	vp_dict["K-Feldspar"] = kspar_vp
	vp_dict["Muscovite (Parallel)"] = msvtpar_vp
	vp_dict["Magnetite"] = mgt_vp
	vp_dict["Pyrite"] = py_vp

	#Create a Vs database
	qtz_vs = 4200
	py_vs = 4950
	pomag_vs = 2720
	kspar_vs = 3375
	mgt_vs = 4200
	msvtpar_vs = 2950
	vs_dict = {}
	vs_dict["Magnetite".lower()] = mgt_vs
	vs_dict["K-Feldspar"] = kspar_vs
	vs_dict["Pyrrhotite (Magnetic)".lower()] = pomag_vs
	vs_dict["Pyrite".lower()] = py_vs
	vs_dict["Muscovite (Parallel)".lower()] = msvtpar_vs
	vs_dict["Quartz".lower()] = qtz_vs

	#WRITE OUT TEST DATABASES
	print "Creating databases...",

	#time to write out appropriate database files
	#hopefully we have write access here!
	#write out a dummy csy file to the correct file
	csy_file = "test.csy"
	write_out_test_csy(csy_file)
	
	#density database
	den_db_name = "den.txt"
	write_out_single_property_db(den_dict, den_db_name)

	#magsus database
	magsus_db_name = "magsus.txt"
	write_out_single_property_db(magsus_dict, magsus_db_name)

	#velocity databases
	vp_db_name = "vp.txt"
	vs_db_name = "vs.txt"
	write_out_single_property_db(vp_dict, vp_db_name)
	write_out_single_property_db(vs_dict, vs_db_name)

	print "... done!"

	#IMPORT TEST DATABASES
	print "Reading in databases...",
	#let's import the CSY file
	molar_volume,molar_mass = csy_reader(csy_file)

	#We can import the density database and verify it works
	den_database = import_single_property_db(den_db_name)	

	#We can also import the magsus database and verify it works
	magsus_database = import_single_property_db(magsus_db_name)

	#We can now import the seismic velocity databases and verify they work
	vp_database = import_single_property_db(vp_db_name)
	vs_database = import_single_property_db(vs_db_name)

	print "... done!"
	print ""

	#PERFORM SOME RESULTS, COMPARE TO HAND-CALCULATED VALUES
	print "Test results:"

	#Calculate some densities
	print "Density"
	print "----------------------"
	#First we calculate by linear average
	#now we can calculate by chemistry
	density_la = calculate_density(mineralogy=mineralogy, type=linear_average, density_db=den_database)
	density_la_hand = qtz_proportion * qtz_den + py_proportion * py_den + pomag_proportion * pomag_den + kspar_proportion * kspar_den + mgt_proportion * mgt_den + msvtpar_proportion * msvtpar_den

	#to calculate density via chemistry, we need to know the num of mols of the materials
	mineralogy_mols = {}
	mineralogy_mols["quartz"] = qtz_proportion / molar_volume["quartz"]
	mineralogy_mols["pyrite"] = py_proportion / molar_volume["pyrite"]
	mineralogy_mols["pyrrhotite (magnetic)"] = pomag_proportion / molar_volume["pyrrhotite (magnetic)"]
	mineralogy_mols["k-feldspar"] = kspar_proportion / molar_volume["k-feldspar"]
	mineralogy_mols["magnetite"] = mgt_proportion / molar_volume["magnetite"]
	mineralogy_mols["muscovite (parallel)"] = msvtpar_proportion / molar_volume["muscovite (parallel)"]
	density_chem = calculate_density(mineralogy=mineralogy_mols, type=from_chemistry, molar_mass_db=molar_mass, molar_volume_db=molar_volume)
	#rather than labouriously calculate it by hand here, I've done it externally
	density_chem_hand = 3.3156500342
	print result_string("Linear average:\t", density_la, density_la_hand)
	print result_string("From chemistry:\t", density_chem, density_chem_hand)

	print ""

	#Magnetic susceptibility test
	print "Magnetic susceptibility"
	print "----------------------"
	magsus_linear = calculate_magsus(mineralogy=mineralogy, type=linear_average, magsus_db=magsus_database)
	magsus_linear_by_hand = qtz_proportion * qtz_magsus + py_proportion * py_magsus + pomag_proportion * pomag_magsus + kspar_proportion * kspar_magsus + mgt_proportion * mgt_magsus + msvtpar_proportion * msvtpar_magsus
	magsus_log = calculate_magsus(mineralogy=mineralogy, type=log_log_average, magsus_db=magsus_database)
	magsus_log_by_hand = math.pow(10,math.log10(py_proportion) * math.log10(py_magsus) + math.log10(pomag_proportion) * math.log10(pomag_magsus) + math.log10(kspar_proportion) * math.log10(kspar_magsus) + math.log10(mgt_proportion) * math.log10(mgt_magsus) + math.log10(msvtpar_proportion) * math.log10(msvtpar_magsus))
	print result_string("Linear average:\t\t  ", magsus_linear, magsus_linear_by_hand)
	print result_string("LogLog average:\t\t  ", magsus_log, magsus_log_by_hand)
	print result_string("Werner (1945):\t\t  ", calculate_magsus(mineralogy=mineralogy, type=werner_1945, magsus_db=magsus_database), 90.0225115)
	#The following methods calculate in SI units, we'll convert to 10^-3 SI.
	print result_string("Mooney & Bleifuss (1953): ", calculate_magsus(mineralogy=mineralogy, type=mooney_bleifuss_1953, magsus_db=magsus_database) * 1000, 284.594224341)
	print result_string("Grant & West (1965):\t  ", calculate_magsus(mineralogy=mineralogy, type=grant_west_1965, magsus_db=magsus_database) * 1000, 220.185740752)
	print result_string("Parasnis (1973):\t  ", calculate_magsus(mineralogy=mineralogy, type=parasnis_1973, magsus_db=magsus_database) * 1000, 66.0916394791)
	print ""

	#Seismic velocity test
	#Let's calculate Vp to begin with...
	vp_linear = calculate_seismic_velocity(mineralogy=mineralogy, type=linear_average, properties=prop_vp, vp_database=vp_database)
	vp_timeavg = calculate_seismic_velocity(mineralogy=mineralogy, type=time_average, properties=prop_vp, vp_database=vp_database)
	#...and calculate it by hand...
	vp_lin_hand = qtz_proportion * qtz_vp + py_proportion * py_vp + pomag_proportion * pomag_vp + kspar_proportion * kspar_vp + mgt_proportion * mgt_vp + msvtpar_proportion * msvtpar_vp
	vp_ta_hand = qtz_proportion / qtz_vp + py_proportion / py_vp + pomag_proportion / pomag_vp + kspar_proportion / kspar_vp + mgt_proportion / mgt_vp + msvtpar_proportion / msvtpar_vp
	vp_ta_hand = 1/vp_ta_hand

	print "Vp"
	print "----------------------"
	print result_string("Linear average:\t", vp_linear, vp_lin_hand)
	print result_string("Time average:\t", vp_timeavg, vp_ta_hand)
	print ""

	#... now Vs...
	vs_linear = calculate_seismic_velocity(mineralogy=mineralogy, type=linear_average, properties=prop_vs, vs_database=vs_database)
	vs_timeavg = calculate_seismic_velocity(mineralogy=mineralogy, type=time_average, properties=prop_vs, vs_database=vs_database)
	#...and calculate it by hand...
	vs_lin_hand = qtz_proportion * qtz_vs + py_proportion * py_vs + pomag_proportion * pomag_vs + kspar_proportion * kspar_vs + mgt_proportion * mgt_vs + msvtpar_proportion * msvtpar_vs
	vs_ta_hand = qtz_proportion / qtz_vs + py_proportion / py_vs + pomag_proportion / pomag_vs + kspar_proportion / kspar_vs + mgt_proportion / mgt_vs + msvtpar_proportion / msvtpar_vs
	vs_ta_hand = 1/vs_ta_hand

	print "Vs"
	print "----------------------"
	print result_string("Linear average:\t", vs_linear, vs_lin_hand)
	print result_string("Time average:\t", vp_timeavg, vp_ta_hand)
	print ""

	#... now Vp/Vs...
	vpvs_linear = calculate_seismic_velocity(mineralogy=mineralogy, type=linear_average, properties=prop_vpvs, vp_database=vp_database, vs_database=vs_database)
	vpvs_timeavg = calculate_seismic_velocity(mineralogy=mineralogy, type=time_average, properties=prop_vpvs, vp_database=vp_database, vs_database=vs_database)

	print "Vp/Vs"
	print "----------------------"
	print result_string("Linear average:\t", vpvs_linear, vp_lin_hand / vs_lin_hand)
	print result_string("Time average:\t", vpvs_timeavg, vp_ta_hand / vs_ta_hand)
	print ""

	#... now we can calculate all 3...
	all_linear = calculate_seismic_velocity(mineralogy=mineralogy, type=linear_average, properties=prop_all, vp_database=vp_database, vs_database=vs_database)
	all_timeavg = calculate_seismic_velocity(mineralogy=mineralogy, type=time_average, properties=prop_all, vp_database=vp_database, vs_database=vs_database)
	all_linear_by_hand = [vp_lin_hand, vs_lin_hand, vp_lin_hand / vs_lin_hand]
	all_timeavg_by_hand = [vp_ta_hand, vs_ta_hand, vp_ta_hand / vs_ta_hand]

	print "All"
	print "----------------------"	
	print result_string_for_lists("Linear average:\t", all_linear, all_linear_by_hand)
	print result_string_for_lists("Time average:\t", all_timeavg, all_timeavg_by_hand)
	print ""


	#clean up - delete those files
	print "Deleting temporary files...",
	os.remove(den_db_name)
	os.remove(magsus_db_name)
	os.remove(vp_db_name)
	os.remove(vs_db_name)
	os.remove(csy_file)
	print "...done!"
	

#if we've called this as a python script for unit testing...
if __name__ == "__main__":
	main()



#end of property_calcs
